// This software including the source code and documentation (the "SOFTWARE")
// is copyrighted.  You are free to use the SOFTWARE for any non-commercial
// purpose whatsoever provided you do not remove this copyright notice.
// If you wish to use the SOFTWARE for commercial purposes, please contact us
// to obtain a license.  The SOFTWARE is provided "as is", with absolutely no
// warranty.  Use it at your own risk.
// Website: http://www.math.ie/checker.html

#include "checker.h"
#include "unique.h"
#include "config.h"
#include <string>
#include <fstream>
#include <time.h>
#include <algorithm>
using namespace std;

#define ZeroMemory( array ) memset( (array), 0, sizeof( array ) );

// we record n-puzzles and pseudo-n-puzzles (i.e. puzzles with
// 2 completions) found, but only up to MAX_PUZZLES of them
int const MAX_PUZZLES = 500;

ClueChecker::ClueChecker( Grid const& grid, char const* pszPuzzFile, int cClues,
                          int nProbability ) : m_grid( grid ), m_pszPuzzFile( pszPuzzFile ),
     m_cClues( cClues ), m_nProbability( nProbability ), m_pTuple( new Cell[ m_cClues ] ),
     m_szPuzzle( new char[ 82 ] ), m_piCandidateClues( new int[ 81 ] )
   {
   // check if there is any checking to do at all (if there are more
   // unavoidable sets than clues, then we cannot possibly get a valid
   // Sudoku puzzle)
   if ( m_cClues < grid.GetNumberOfDisj() )
      {
      printf( "\nThere does not exist a %d as we have found %d disjoint "
              "unavoidable sets.\n", m_cClues, grid.GetNumberOfDisj() );
      exit( 0 );
      }
   // allow for a progress indicator
   // the following three lines mean: m_cProgressRounds = min( grid.GetNumberOfDisj() - 1, 6 )
   m_cProgressRounds = grid.GetNumberOfDisj() - 1;
   if ( m_cProgressRounds > 6 )
      m_cProgressRounds = 6;

   // initialize random generator
   srand( static_cast< unsigned int >( time( NULL ) ) );

   ComputeHittingMasks();

   // set hitting masks to 0
   memset( &m_SetMask, 0, sizeof( m_SetMask ) );

   // set mask of dead clues to 0
   ZeroMemory( m_aiDeadClues );
   m_cPuzzles        = 0;
   m_cPuzzlesSolved  = 0;
   m_cRounds         = 0;
   m_cUniques        = 0;
   m_cPseudos        = 0;
   m_iIndex          = 0;
   m_bRun = false; // this is the first time we run (see HasRunBefore())

   // initialise m_aiLowestBit
   m_aiLowestBit = new int[ 65536 ];
   for ( int i = 0;  i < 65536;  i++ )
      {
      for ( int j = 0;  j <= 16;  j++ )
         if ( (i & (1 << j)) == 0 )
            {
            m_aiLowestBit[ i ] = j;
            break;
            }
      }
   }


ClueChecker::~ClueChecker()
   {
   delete m_pTuple;
   delete m_szPuzzle;
   delete m_piCandidateClues;
   }


void ClueChecker::DoRandomChecking()
   {
   if ( HasRunBefore() )
      return;

   if ( m_nProbability > 0 )
      printf( "Beginning a random search for a %d now, solving %d%% of puzzles ...\n",
              m_cClues, m_nProbability );
   else
      printf( "Estimating number of puzzles to solve when searching for a %d by searching randomly ...\n",
              m_cClues );

   // this has to be less than or equal to the MCN
   int cCluesToPick = m_cProgressRounds + 1;
   // compute total number of tuples of size cCluesToPick
   int cMaxTuples = 1;
   for ( int j = 0;  j < cCluesToPick;  j++ )
      cMaxTuples *= m_grid.GetUnav()[ j ].size();
   // get memory so that we can store which tuple we've checked already
   // (to ensure that each tuple is checked only once)
   char* anTuples = new char[ cMaxTuples ];
   memset( anTuples, 0, cMaxTuples * sizeof( char ) );

   // open files to keep track of (pseudo-)puzzles found
   // first "calculate" filenames
   char sz[ 120 ], szPseudo[ 120 ];
   sprintf( sz, "-%d-puzzles." DATA_FILE_EXTN, m_cClues );
   sprintf( szPseudo, "-pseudo-%d-puzzles." DATA_FILE_EXTN, m_cClues - 1 );
   // open files
   std::string sFilePuzz( m_pszPuzzFile );
   sFilePuzz += sz;
   m_posPuzzFile = new ofstream( sFilePuzz.c_str() );
   std::string sFilePseuPuzz( m_pszPuzzFile );
   sFilePseuPuzz += szPseudo;
   if ( DoWeWantPseudos() )
      m_posPseuPuzzFile = new ofstream( sFilePseuPuzz.c_str() );
   else
      m_posPseuPuzzFile = NULL;

   // begin enumeration
   m_Stopwatch.Start();
   // counter
   int cTuplesChecked = 0;
   // check up to 25% of tuples randomly
   while ( 4 * cTuplesChecked < cMaxTuples )
      {
      // generate random tuple
      m_iIndex = 0;
      // set mask of dead clues to 0
      ZeroMemory( m_aiDeadClues );
      // set hitting masks to 0
      memset( &m_SetMask, 0, sizeof( m_SetMask ) );
      int nTupleNumber = 0;
      for ( int iSet = 0;  iSet < cCluesToPick;  iSet++ )
         {
         int iClue = rand() % m_grid.GetUnav()[ iSet ].size();
         nTupleNumber *= m_grid.GetUnav()[ iSet ].size();
         nTupleNumber += iClue;
         // AddClue will increment m_iIndex
         AddClue( iSet, iClue );
         }

      // make sure we haven't had this tuple before
      if ( anTuples[ nTupleNumber ] == 0 )
         {
         if ( m_iIndex < m_grid.GetNumberOfDisj() )
            GenerateTuples();
         else
            ProcessTuple();
         anTuples[ nTupleNumber ] = 1; // mark this tuple as processed
         cTuplesChecked++;

         // status display
         float flSec = m_Stopwatch.GetTimeElapsed() * cMaxTuples / cTuplesChecked;
         if ( DoWeWantPseudos() )
            {
#ifdef _WIN32
         printf( "%.3f%% of tuples checked randomly, %I64d puzzles gen'd\n"
                 "   %d uniques, %d pseudos found; ET for full search: ",
#else
         printf( "%.3f%% of tuples checked randomly; %lld puzzles gen'd\n"
                 "   %d uniques, %d pseudos found; ET for full search: ",
#endif
                 static_cast< float >( cTuplesChecked ) / cMaxTuples * 100, m_cPuzzles,
                 m_cUniques, m_cPseudos );
            }
         else
            {
#ifdef _WIN32
         printf( "%.3f%% of tuples checked randomly, %I64d puzzles gen'd\n"
                 "   %d uniques found; ET for full search: ",
#else
         printf( "%.3f%% of tuples checked randomly; %lld puzzles gen'd\n"
                 "   %d uniques found; ET for full search: ",
#endif
                 static_cast< float >( cTuplesChecked ) / cMaxTuples * 100, m_cPuzzles,
                 m_cUniques );
            }
         PrintTime( static_cast< int >( flSec ) );
         printf( "\n" );
         // print time spent in solver, if desired
         if ( DoWeWantStats() )
            {
            if ( m_cPuzzlesSolved > 0 )
               printf( "   Time spent in solver: %.1f%% (%d puzzles / second).\n",
                       m_StopwatchSolver.GetTimeElapsed() / m_Stopwatch.GetTimeElapsed() * 100,
                       ( int )( m_cPuzzlesSolved / m_StopwatchSolver.GetTimeElapsed() + 0.5 ) );
            else
               printf( "   Time spent in solver: n/a (no puzzles run through solver so far).\n" );
            }
         }
      }
   delete m_posPuzzFile;
   delete m_posPseuPuzzFile;

   cout << endl << "Finished searching 25% of tuples randomly.\n"
                   "Total computation time was ";
   PrintTime( static_cast< int >( m_Stopwatch.GetTimeElapsed() ) );
   cout << "." << endl;
   
   printf( "\nAny puzzles or pseudo-puzzles found were saved to\n%s respectively %s.\n",
           sFilePuzz.c_str(), sFilePseuPuzz.c_str() );
   }


bool ClueChecker::HasRunBefore()
   {
   // make sure we haven't run before (o'wise, some of the initialisation
   // that currently takes place in the constructor has to be moved to
   // this method and to ClueChecker::DoChecking)
   if ( m_bRun )
      {
      cout << "Error: must re-initalise certain member variables (requires "
           << endl << "modifications to ClueChecker).";
      return true;
      }
   m_bRun = true;
   return false;
   }


void ClueChecker::DoChecking( int cFirstClues /* = 0 */ )
   {
   if ( HasRunBefore() )
      return;

   if ( cFirstClues > 0 )
      {
      // have the user select a clue from the first unavoidable set
      cout << endl << "You have indicated that you wish to specify the first "
                   << cFirstClues << " clues by hand." << endl;
      if ( cFirstClues > m_grid.GetNumberOfDisj() )
         {
         cFirstClues = m_grid.GetNumberOfDisj();
         cout << "Note: the maximum number of clues you can specify is " << cFirstClues <<
                 " for this grid" << endl << "(=the MCN of this grid)." << endl;
         }

      // now have the user specify clues one at a time
      while ( m_iIndex < cFirstClues )
         {
         UnavSet const& set = m_grid.GetUnav()[ m_iIndex ];
         cout << "Please choose a clue from the following set:" << endl;
         set.Print();
         cout << endl;

         // get selection
         int iCell = -1;
         while ( iCell == -1 )
            {
            int iClue;
            cout << "Enter clue: ";
            cin >> iClue;
            // find the index of the clue in this set
            for ( int i = 0;  i < m_grid.GetUnav()[ m_iIndex ].size();  i++ )
               if ( set[ i ] == iClue )
                  iCell = i;

            if ( iCell == -1 )
               cout << iClue << " is not a valid clue.  "
                                "Please choose from the above only." << endl;
            }
         cout << endl << "Using " << set[ iCell ] <<
                 " as clue no. " << m_iIndex + 1 << "." << endl << endl;

         // add clue to tuple
         AddClue( m_iIndex, iCell );
         }
      }

   if ( m_nProbability > 0 )
      printf( "Beginning search for a %d now, solving %d%% of puzzles ...\n",
              m_cClues, m_nProbability );
   else
      printf( "Counting number of puzzles to solve when searching for a %d ...\n",
              m_cClues );

   // initialize counter
   m_cTotalRounds = 1;
   for ( int j = cFirstClues;  j < m_cProgressRounds;  j++ )
      m_cTotalRounds *= m_grid.GetUnav()[ j ].size();

   // open files to keep track of (pseudo-)puzzles found
   // first "calculate" filenames
   char sz[ 120 ], szPseudo[ 120 ];
   sprintf( sz, "-%d-puzzles", m_cClues );
   sprintf( szPseudo, "-pseudo-%d-puzzles", m_cClues - 1 );
   for ( int i = 0;  i < cFirstClues;  i++ )
      {
      sprintf( sz + strlen( sz ), "-%d", m_pTuple[ i ].m_Coords );
      sprintf( szPseudo + strlen( szPseudo ), "-%d", m_pTuple[ i ].m_Coords );
      }
   strcat( sz, "." DATA_FILE_EXTN );
   strcat( szPseudo, "." DATA_FILE_EXTN );
   // open files
   std::string sFilePuzz( m_pszPuzzFile );
   sFilePuzz += sz;
   m_posPuzzFile = new ofstream( sFilePuzz.c_str() );
   std::string sFilePseuPuzz( m_pszPuzzFile );
   sFilePseuPuzz += szPseudo;
   if ( DoWeWantPseudos() )
      m_posPseuPuzzFile = new ofstream( sFilePseuPuzz.c_str() );
   else
      m_posPseuPuzzFile = NULL;

   // begin enumeration
   m_Stopwatch.Start();
   GenerateTuples();
   m_Stopwatch.Pause();
   delete m_posPuzzFile;
   delete m_posPseuPuzzFile;

   cout << endl << "Finished.  Total computation time was ";
   PrintTime( static_cast< int >( m_Stopwatch.GetTimeElapsed() ) );
   cout << "." << endl;
   
   if ( cFirstClues > 0 )
      {
      cout << "Only a partial search was done -- searched the branch whose first clues are ";
      for ( int i = 0;  i < cFirstClues;  i++ )
         {
         cout << m_pTuple[ i ].m_Coords;
         if ( i != cFirstClues - 1 )
            cout << ", ";
         }
      cout << endl;

      if ( m_nProbability == 100 ) // all puzzles solved
         {
         // found any n-puzzles?
         if ( m_cUniques == 0 )
            {
            printf( "No %d-puzzle exists for this grid in this branch (all puzzles were "
                    "solved).\n", m_cClues );

            if ( m_cPseudos == 0 )           
               printf( "No pseudo-%d-puzzle exists for this grid in this branch (all puzzles "
                       "were\nsolved).\n", m_cClues - 1 );
            else
               {
               printf( "There are %d pseudo-%d-puzzles for this grid in this branch (all "
                       "puzzles\nwere solved).\n",
                       m_cPseudos, m_cClues - 1 );
               }
            }
         else
            {
            printf( "There are %d %d-puzzles in this branch (all puzzles were solved).\n",
                    m_cUniques, m_cClues );
            }
         }
      else if ( m_nProbability > 0 )
         {
         if ( m_cUniques == 0 )
            {
            printf( "No %d-puzzle found for this grid in this branch (but only solved %d%% "
                    "of\npuzzles).\n", m_cClues, m_nProbability );
            if ( m_cPseudos == 0 )           
               printf( "No pseudo-%d-puzzle exists for this grid in this branch (but only "
                       "solved %d%% of\npuzzles).\n", m_cClues - 1, m_nProbability );
            else
               {
               printf( "Found %d pseudo-%d-puzzles for this grid in this branch (but only\n"
                       "solved %d%% of puzzles).", m_cPseudos, m_cClues - 1, m_nProbability );
               }
            }
         else
            {
            printf( "Found %d %d-puzzles for this grid in this branch (but only solved\n"
                    "%d%% of puzzles).\n", m_cUniques, m_cClues, m_nProbability );
            }
         }
      else
         printf( "Total number of puzzles to be solved in this branch (puzzles were only\n"
                 "generated): %d.\n", m_cPuzzles );
      }
   else
      {
      cout << "A full search was done (all puzzles were enumerated)." << endl << endl;
      if ( m_nProbability == 100 ) // all puzzles solved
         {
         // found any n-puzzles?
         if ( m_cUniques == 0 )
            {
            printf( "No %d-puzzle exists for this grid (all puzzles were solved).\n",
                    m_cClues );

            if ( m_cPseudos == 0 )           
               printf( "No pseudo-%d-puzzle exists for this grid (all puzzles were solved).\n",
                       m_cClues - 1 );
            else
               {
               printf( "There are %d pseudo-%d-puzzles for this grid (all puzzles were "
                       "solved).\n", m_cPseudos, m_cClues - 1 );
               }
            }
         else
            {
            printf( "There are %d %d-puzzles for this grid (all puzzles were solved).\n",
                    m_cUniques, m_cClues );
            }
         }
      else if ( m_nProbability > 0 )
         {
         if ( m_cUniques == 0 )
            {
            printf( "No %d-puzzle found for this grid (but only solved %d%% "
                    "of puzzles).\n", m_cClues, m_nProbability );
            if ( m_cPseudos == 0 )           
               printf( "No pseudo-%d-puzzle exists for this grid (but only "
                       "solved %d%% of puzzles).\n", m_cClues - 1, m_nProbability );
            else
               {
               printf( "Found %d pseudo-%d-puzzles for this grid (but only solved\n"
                       "%d%% of puzzles).\n", m_cPseudos, m_cClues - 1, m_nProbability );
               }
            }
         else
            {
            printf( "Found %d %d-puzzles for this grid (but only solved %d%% of puzzles).\n",
                    m_cUniques, m_cClues, m_nProbability );
            }
         }
      else
         printf( "Total number of puzzles to be solved (puzzles were only generated): %d.\n",
                 m_cPuzzles );
      }
   printf( "\nAny puzzles or pseudo-puzzles found were saved to\n%s respectively %s.\n",
           sFilePuzz.c_str(), sFilePseuPuzz.c_str() );
   }


// generate all tuples based on our collection of disjoint unavoidable sets
// for each such tuple, add clues first on the basis of any remaining
// unavoidable sets that are not hit yet, finally add the final clues by
// trying all possibilities
void ClueChecker::GenerateTuples()
   {
   // print progress update every now and then
   if ( m_iIndex == m_cProgressRounds )
      {
      if ( m_cRounds > 0 )
         {
         // obtain total time passed so far and compute ETTF
         float flSec = m_Stopwatch.GetTimeElapsed() * (m_cTotalRounds - m_cRounds) / m_cRounds;
         // status display
         if ( DoWeWantPseudos() )
            {
#ifdef _WIN32
         printf( "%d/%d (%I64d puzzles gen'd; %d uniques, %d pseudos found); ETTF ",
#else
         printf( "%d/%d (%lld puzzles gen'd; %d uniques, %d pseudos found); ETTF ",
#endif
                 m_cRounds, m_cTotalRounds, m_cPuzzles, m_cUniques, m_cPseudos );
            }
         else
            {
#ifdef _WIN32
         printf( "%d/%d (%I64d puzzles gen'd; %d uniques found); ETTF ",
#else
         printf( "%d/%d (%lld puzzles gen'd; %d uniques found); ETTF ",
#endif
                 m_cRounds, m_cTotalRounds, m_cPuzzles, m_cUniques );
            }
         PrintTime( static_cast< int >( flSec ) );
         printf( "\n" );

         // print time spent in solver, if desired
         if ( DoWeWantStats() )
            {
            printf( "   Time spent in solver: %.1f%% (%d puzzles / second).\n",
                    m_StopwatchSolver.GetTimeElapsed() / m_Stopwatch.GetTimeElapsed() * 100,
                    ( int )( m_cPuzzlesSolved / m_StopwatchSolver.GetTimeElapsed() + 0.5 ) );
            }
         }
      m_cRounds++;
      }

   UnavSet const& set = m_grid.GetUnav()[ m_iIndex ];
   for ( int i = 0;  i < set.size();  i++ )
      {
      int aiDeadCluesOld[ 3 ];
      memcpy( aiDeadCluesOld, m_aiDeadClues, sizeof( m_aiDeadClues ) );
      bitmask_t OldSetMask = m_SetMask;
      AddClue( m_iIndex, i ); // this increments m_iIndex by 1
      if ( m_iIndex < m_grid.GetNumberOfDisj() )
         {
         // there are more disjoint unavoidable sets we need to hit
         GenerateTuples();
         }
      else
         {
         // have one clue from each of the disjoint unavoidable sets
         // however, not all of the remaining unavoidable sets may be hit yet;
         // use this information to draw more clues
         ProcessTuple();
         }
      m_SetMask = OldSetMask;
      m_iIndex--;
      memcpy( m_aiDeadClues, aiDeadCluesOld, sizeof( m_aiDeadClues ) );
      }
   }


void ClueChecker::ProcessTuple()
   {
   // iNotHit is passed by reference and gives the index of the
   // first unavoidable set that is not hit by the g_tuple
   // start searching from the next set after the one we drew the previus clue from
   // (the unavoidable sets with smaller indices are already hit)
   int iNotHit;
   if ( ! AreAllUnavoidableSetsHit( iNotHit ) )
      {
      // good: not yet all unavoidable sets hit (iNotHit now contains the
      // index of the first unavoidable set not hit by our current tuple)

      if ( m_iIndex == m_cClues )
         {
         // nothing to do: we already have collected enough clues, but still haven't
         // hit all unavoidable sets
         return;
         }

      // now draw the next clue from the first unvoidable set not hit 
      // (do this in all possible ways, of course)
      UnavSet const& set = m_grid.GetUnav()[ iNotHit ];
      for ( int i = 0;  i < set.size();  i++ )
         {
         int idx = Coords2Idx( set[ i ] );
         if ( (m_aiDeadClues[ idx / 32 ] & (1 << (idx % 32))) == 0 )
            {
            // clue is not dead yet, so try this possibility
            int aiDeadCluesOld[ 3 ];
            memcpy( aiDeadCluesOld, m_aiDeadClues, sizeof( m_aiDeadClues ) );
            bitmask_t OldSetMask = m_SetMask;
            AddClue( iNotHit, i ); // this increments m_iIndex by 1
            ProcessTuple();
            m_SetMask = OldSetMask;
            m_iIndex--;
            memcpy( m_aiDeadClues, aiDeadCluesOld, sizeof( m_aiDeadClues ) );
            }
         }
      }
   else
      {
      if ( m_iIndex == m_cClues )
         {
         // check puzzle for a unique solution
         GenerateAndCheckPuzzle();
         }
      else
         {
         // still need to draw more clues
         // first make a list of cells that are 'out' i.e. cells that are
         // not candidates for the remaining clues to draw
         // these include of course all the clues we've already used
         // in addition, since the order of clues is irrelevant, we assume
         // that if a clue comes from unavoidable set with index i, any
         // further clues drawn from unavoidable set i are greater
         // (greater = higher numerical order)
         bool abCellUsed[ 81 ];
         memset( abCellUsed, 0, sizeof( abCellUsed ) ); // set all elements to 'false'
         for ( int k = 0;  k < m_iIndex;  k++ )
            {
            UnavSet const& set = m_grid.GetUnav()[ m_pTuple[ k ].m_iSet ];
            for ( int l = 0;  l <= m_pTuple[ k ].m_iCell;  l++ )
               abCellUsed[ Coords2Idx( set[ l ] ) ] = true;
            }

         // now assemble the remaining cells into a single string
         int cCell = 0;
         m_cCandidateClues = 0;
         for ( int i = 1;  i <= 9;  i++ )
            for ( int j = 1;  j <= 9;  j++ )
               if ( ! abCellUsed[ cCell++ ] )
                  m_piCandidateClues[ m_cCandidateClues++ ] = i*10 + j;

         DrawRemainingClues();
         }
      }
   }


inline void ClueChecker::GenerateAndCheckPuzzle()
   {
   m_cPuzzles++;
   if ( m_nProbability < 100  &&  (rand() % 100) >= m_nProbability )
      return;
   // set up puzzle
   memset( m_szPuzzle, '0', 81 );   // fill string with 81 0's (the digit '0')
   m_szPuzzle[ 81 ] = 0;            // terminate string with a zero byte
   // place clues according to the tuple just generated
   for ( int i = 0;  i < m_cClues;  i++ )
      {
      int ich = Coords2Idx( m_pTuple[ i ].m_Coords );
      m_szPuzzle[ ich ] = m_grid.GetGrid()[ ich ];
      }

   // run puzzle through solver, measure how long this takes
   m_StopwatchSolver.Start();
   int cCompletions = GetNumberOfCompletionsOfPuzzle( m_szPuzzle );
   m_cPuzzlesSolved++;
   m_StopwatchSolver.Pause();
   if ( cCompletions == 1 )
      {
      // make a copy of the n-puzzle, but record at most MAX_PUZZLES
      Puzzle puzzle( m_cClues, m_pTuple, cCompletions );
      if ( cCompletions == 1 )         
         {
         if ( m_vcUniquePuzzles.size() < MAX_PUZZLES  &&
              m_vcUniquePuzzles.end() == find( m_vcUniquePuzzles.begin(),
                                               m_vcUniquePuzzles.end(), puzzle ) )
            {
            m_cUniques++;
            m_vcUniquePuzzles.push_back( puzzle );
            puzzle.Print( *m_posPuzzFile, m_grid );
            m_posPuzzFile->flush();
            }
         }

      // next see if this n-puzzle gives rise to any pseudo-(n-1)-puzzles (i.e.
      // (n-1)-puzzles with exactly 2 completions), if the user is interested in those
      if ( DoWeWantPseudos() )
         {
         // this enabling-disabling is done to speed things up; we distinguish
         // between puzzles having 2 resp. 3+ solutions only when we actually need to
         EnablePseudos();
         // make a copy of n-puzzle, take all (n-1)-subpuzzles and see if any of them
         // has exactly 2 completions
         char* psz = new char[ 81 ];
         memcpy( psz, m_szPuzzle, 81 );
         int i = 0;
         do
            {
            // search for next clue, stop if reached past the end of the string
            while ( i < 81  &&  psz[ i ] == '0' )
               i++;
            if ( i == 81 )
               break;

            // delete clue
            char iOldCell = psz[ i ];
            psz[ i ] = '0';
            cCompletions = GetNumberOfCompletionsOfPuzzle( psz );
            if ( cCompletions == 2 )
               {
               Puzzle puzzPseudo( psz, 2 );
               if ( m_vcPseudoPuzzles.size() < MAX_PUZZLES  &&
                    m_vcPseudoPuzzles.end() == find( m_vcPseudoPuzzles.begin(),
                                                     m_vcPseudoPuzzles.end(), puzzPseudo ) )
                  {
                  m_cPseudos++;
                  m_vcPseudoPuzzles.push_back( puzzPseudo );
                  puzzPseudo.Print( *m_posPseuPuzzFile, m_grid );
                  m_posPseuPuzzFile->flush();
                  }
               }
            // restore the clue we deleted above and move to next clue
            psz[ i++ ] = iOldCell;
            }
         while ( true );

         delete[] psz;
         DisablePseudos();
         }
      }
   else if ( cCompletions <= 0 )
      {
      printf( "cCompletions = %d\n", cCompletions );
      // each puzzle has at least one completion, otherwise there is a mistake somewhere
      printf( "\nSomething's wrong -- found a puzzle with no solutions or an inavlid puzzle "
              "(this\nmeans there is a mistake in this program -- please let us know):\n%s\n",
              m_szPuzzle );
      exit( 0 );
      }
   }


// check if our current tuple hits all unavoidable sets (starting from
// the set with index iNotHit)
// if all sets are hit the function returns true;
// otherwise it returns false and iNotHit is the index of the
// first unavoidable set that is not hit
inline bool ClueChecker::AreAllUnavoidableSetsHit( int& iNotHit )
   {
   int64 nSetMask64;
   int bAdd;
   // first find out which 64-bit integer is the one
   if ( m_SetMask.aBits[ 0 ] + 1 == 0LL )
      {
#if NUM_PACKS > 1
      if ( m_SetMask.aBits[ 1 ] + 1 == 0LL )
         {
#if NUM_PACKS > 2
         if ( m_SetMask.aBits[ 2 ] + 1 == 0LL )
            {
#if NUM_PACKS > 3
            if ( m_SetMask.aBits[ 3 ] + 1 == 0LL )
               {
#if NUM_PACKS > 4
               if ( m_SetMask.aBits[ 4 ] + 1 == 0LL )
                  return true;
               else
                  {
                  nSetMask64 = m_SetMask.aBits[ 4 ];
                  bAdd = 256;
                  }
#else
               return true;
#endif
               }
            else
               {
               nSetMask64 = m_SetMask.aBits[ 3 ];
               bAdd = 192;
               }
#else
            return true;
#endif
            }
         else
            {
            nSetMask64 = m_SetMask.aBits[ 2 ];
            bAdd = 128;
            }
#else
         return true;
#endif
         }
      else
         {
         nSetMask64 = m_SetMask.aBits[ 1 ];
         bAdd = 64;
         }
#else
      return true;
#endif
      }
   else
      {
      nSetMask64 = m_SetMask.aBits[ 0 ];
      bAdd = 0;
      }

   // test least significant 32 bit
   int nSetMask32;
   if ( static_cast< int >( nSetMask64 ) == 0xFFFFFFFF )
      {
      nSetMask32 = static_cast< int >( nSetMask64 >> 32 );
      bAdd += 32;
      }
   else
      nSetMask32 = static_cast< int >( nSetMask64 );

   if ( (nSetMask32 & 0xFFFF) == 0xFFFF )
      {
      nSetMask32 >>= 16;
      bAdd += 16;
      }
   // Clear upper 16 bits.  This is always necessary because nSetMask32
   // might be negative.
   // Example. int x = 0x80000000; x >>= 16; // now x = 0xFFFF8000;
   nSetMask32 &= 0xFFFF;
   iNotHit = m_aiLowestBit[ nSetMask32 ];

   // finally, also count sets from max clique as well as the
   // 'offset' computed above
   iNotHit += m_grid.GetNumberOfDisj() + bAdd;
   return ( iNotHit == m_grid.GetNumberOfUnav() );
   }


inline void ClueChecker::DrawRemainingClues( int iStart /* = 0 */ )
   {
   if ( m_iIndex == m_cClues - 1 )
      {
      // draw final clue
      for ( int i = iStart;  i < m_cCandidateClues;  i++ )
         {
         m_pTuple[ m_iIndex ].m_Coords = m_piCandidateClues[ i ];
         GenerateAndCheckPuzzle();
         }
      }
   else
      {
      // not yet the last clue we're drawing (i.e. have to draw more clues than
      // just this one)
      for ( int i = iStart;  i < m_cCandidateClues - (m_cClues - m_iIndex - 1);  i++ )
         {
         m_pTuple[ m_iIndex ].m_Coords = m_piCandidateClues[ i ];
         m_iIndex++;
         DrawRemainingClues( i + 1 );
         m_iIndex--;
         }
      }
   }


void ClueChecker::ComputeHittingMasks()
   {
   int cDisj = m_grid.GetNumberOfDisj();
   for ( int i = 1;  i <= 9;  i++ )
      for ( int j = 1;  j <= 9;  j++ )
         {
         int nCellCoords = 10*i + j;
         memset( &(m_aHittingMasks[ nCellCoords ]), 0, sizeof( bitmask_t ) );
         for ( int iSet = cDisj;  iSet < m_grid.GetNumberOfUnav();  iSet++ )
            {
            bool bFound = false;
            for ( int k = 0;  k < m_grid.GetUnav()[ iSet ].size()  &&  ! bFound;  k++ )
               if ( m_grid.GetUnav()[ iSet ][ k ] == nCellCoords )
                  bFound = true;
            if ( bFound )
               {
               int iBit = iSet - cDisj;
               m_aHittingMasks[ nCellCoords ].aBits[ iBit / 64 ] |= ((int64)1) << (iBit % 64); // 2^(iSet - cDisj)
               }
            }
         }
   }


// add a clue "manually" to the current tuple
inline void ClueChecker::AddClue( int iSet, int iCell )
   {
   UnavSet const& set = m_grid.GetUnav()[ iSet ];
   // add clue to tuple
   m_pTuple[ m_iIndex ].m_iSet   = iSet;
   m_pTuple[ m_iIndex ].m_iCell  = iCell;
   m_pTuple[ m_iIndex ].m_Coords = set[ iCell ];
   // make a note that the unavoidable sets that containt this clue have been hit
   XOR_BITMASKS( m_pTuple[ m_iIndex ].m_Coords );
   // update bitmask of dead clues
   m_aiDeadClues[ 0 ] |= set.m_aiDeadClues[ iCell ][ 0 ];
   m_aiDeadClues[ 1 ] |= set.m_aiDeadClues[ iCell ][ 1 ];
   m_aiDeadClues[ 2 ] |= set.m_aiDeadClues[ iCell ][ 2 ];
   // advance to next clue
   m_iIndex++;
   }
