// This software including the source code and documentation (the "SOFTWARE")
// is copyrighted.  You are free to use the SOFTWARE for any non-commercial
// purpose whatsoever provided you do not remove this copyright notice.
// If you wish to use the SOFTWARE for commercial purposes, please contact us
// to obtain a license.  The SOFTWARE is provided "as is", with absolutely no
// warranty.  Use it at your own risk.
// Website: http://www.math.ie/checker.html

#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
#include "grid.h"
#include "checker.h"
#include "unavsets.h"
#include "gunav.h"
#include "config.h"
#include "graph.h"

using namespace std;

bool ReadGridFromFile( char const* pszFilename, char* pszGrid );

Grid::Grid( char const* pszFilename, char const* pszFilenameExtraUnavoidables )
   {
   m_pszGrid = new char[ 82 ];
   if ( ! ReadGridFromFile( pszFilename, m_pszGrid ) )
      exit( 0 );

   m_sFileUnav = GetUnavFilename( pszFilename );
   ComputeUnavoidableSets( m_pszGrid, pszFilenameExtraUnavoidables,
                           &m_pUnav, &m_cUnav, &m_cDisjoints );
   }


Grid::~Grid()
   {
   delete m_pszGrid;
   }


bool ReadGridFromFile( char const* pszFilename, char* pszGrid )
   {
   ifstream file( pszFilename );
   if ( ! file )
      {
      cout << "Unable to open file " << pszFilename << "." << endl;
      return false;
      }

   if ( ! file.getline( pszGrid, 82 ) )
      {
      cout << "Error reading from " << pszFilename << "." << endl;
      return false;
      }

   pszGrid[ 81 ] = '\0';
   if ( strlen( pszGrid ) < 81 )
      {
      cout << "Erroneous line encountered (length < 81 )." << endl;
      return false;
      }

   for ( int i = 0;  i < 81;  i++ )
      if ( pszGrid[ i ] < '1'  ||  pszGrid[ i ] > '9' )
         {
         cout << "Illegal character encountered: '" << pszGrid[ i ]
              << "' (only digits 1-9 are allowed)." << endl;
         return false;
         }

   return true;
   }


void Grid::ComputeUnavoidableSets( char const* pszGrid,
                                   char const* pszFilenameExtraUnavoidables,
                                   UnavSet** ppUnav, int* pcUnav, int* pcDisj )
   {
   vector< MySet > sets;
   int size4or6 = 0;
   // first try to read the unavoidables from a file (which should exist if
   // unavoid or checker has been run on this grid before)
   int cSets = ReadAdditionalUnavoidables( m_sFileUnav.c_str(), sets );
   if ( cSets > 0 )
      {
      sort( sets.begin(), sets.end() );
      RemoveDuplicateSets( sets );
      TestSetsForUnavoidability( sets, pszGrid );
      cout << "Read " << ( int ) sets.size() << " unavoidable sets from " <<
               m_sFileUnav.c_str() << " (will not compute\nunavoidable sets again, "
               "please delete the file " << m_sFileUnav.c_str() << " to\ncompute "
               "the unavoidable sets for this grid from scratch)." << endl << endl;
      // count the # of sets of size 4 or 6
      for ( vector< MySet >::iterator it = sets.begin();  it != sets.end();  it++ )
         if ( (*it).GetSize() == 4  ||  (*it).GetSize() == 6 )
            size4or6++;
      }
   else
      {
      if ( ShouldWeFindAllUnavoidables() )
         cout << "Finding unavoidable sets (this takes about two minutes on a 2 GHz computer).\n";
      else
         cout << "Finding only a smaller collection of unavoidable sets (which is quicker).\n";
      int A[ 9 ][ 9 ];
      for ( int i = 0;  i < 81;  i++ )
         A[ i / 9 ][ i % 9 ] = pszGrid[ i ] - '0';

      FindUnavoidableSets( A, sets );

      // now compute the transpose of the matrix and find all unavoidable sets for the transpose
      int A_transpose[ 9 ][ 9 ];
      for ( int i = 0;  i < 9;  i++ )
         for ( int j = 0;  j < 9;  j++ )
            A_transpose[ i ][ j ] = A[ j ][ i ];
      // third parameter = true means swap rows and columns (to un-transpose the coordinates)
      FindUnavoidableSets( A_transpose, sets, true );
      // certain unavoidable sets of size 6 are found twice, remove them now
      RemoveDuplicateSets( sets );
      TestSetsForUnavoidability( sets, pszGrid );
      size4or6 = static_cast< int >( sets.size() );

      // now find unavoidable sets of size >= 8 (using Guenter's implementation)
      int B[ 82 ];
      for ( int i = 0;  i < 81;  i++ )
         B[ i + 1 ] = A[ i / 9 ][ i % 9 ];
      GFindUnavoidableSets( B, sets );

      // find unavoidables of sizes 8 to 12
      if ( ShouldWeFindAllUnavoidables() )
         {
         FindUnavoidablesGeneral( A, sets );
         FindUnavoidablesGeneral( A_transpose, sets, true );
         cout << endl << endl;
         // save unavoidables just found to a file so we don't have to compute them again
         sort( sets.begin(), sets.end() );
         RemoveDuplicateSets( sets );
         TestSetsForUnavoidability( sets, pszGrid );
         SaveUnavoidablesToFile( m_sFileUnav.c_str(), sets );
         }
      }

   // read any additional unavoidable sets (from a file supplied by the user)
   if ( pszFilenameExtraUnavoidables != NULL )
      ReadAdditionalUnavoidables( pszFilenameExtraUnavoidables, sets );
   sort( sets.begin(), sets.end() );
   RemoveDuplicateSets( sets );
   // test if all the candidates found are actually unavoidable, remove those that aren't
   // (for ordinary/classical Sudoku, all sets found *should* be unavoidable, however
   // for variants like Sudoku-X, calling this function is absolutely crucial)
   TestSetsForUnavoidability( sets, pszGrid );
   cout << "Found " << ( int ) sets.size() << " unavoidable sets (" << size4or6
        << " of size 4 or 6)." << endl;

   // set up a graph whoses vertices are the unavoidable sets we have
   // two vertices are adjacent <=> the corresponding sets are disjoint
   // find a maximal clique in this graph
   vector< int > clique;
   Graph g( static_cast< int >( sets.size() ) );
   for ( int i = 0;  i < static_cast< int >( sets.size() );  i++ )
      for ( int j = i;  j < static_cast< int >( sets.size() );  j++ )
         if ( sets[ i ].Disjoint( sets[ j ] ) )
            g.AddEdge( i, j );
   // find a max clique in this graph
   g.FindMaxClique( &clique );

   sort( clique.begin(), clique.end() );
   cout << "The maximum # of disjoint unavoidable sets (max clique number -- MCN) is " <<
           static_cast< int >( clique.size() ) << "." << endl;
   *pcDisj = static_cast< int >( clique.size() );
   // allow only up to NUM_PACKS * 64 additional unavoidables
   *pcUnav = static_cast< int >( min( clique.size() + NUM_PACKS * 64, sets.size() ) );
   if ( *pcUnav < static_cast< int >( sets.size() ) )
      cout << "NOTE: using only " << NUM_PACKS * 64 << " additional unavoidable sets (not counting those\n"
              "unavoidable sets that belong to the max collection of pairwise\n"
              "disjoint unavoidables).  Please use another version of checker\n"
              "if this is not sufficient." << endl;
   *ppUnav = new UnavSet[ *pcUnav ];

   int* anMemberOfDisj = new int[ sets.size() ];
   memset( anMemberOfDisj, 0, sizeof( int ) * sets.size() );
   int iSet = 0;
   for ( vector< int >::iterator it = clique.begin();  it != clique.end();  it++ )
      {
      anMemberOfDisj[ *it ] = 1;
      (*ppUnav)[ iSet++ ].SetData( &sets[ *it ] );
      }

   for ( int i = 0;  i < static_cast< int >( sets.size() )  &&  iSet < *pcUnav;  i++ )
      if ( anMemberOfDisj[ i ] == 0 )
         (*ppUnav)[ iSet++ ].SetData( &sets[ i ] );

   if ( iSet != *pcUnav )
      {
      cout << "There's something wrong ... " << endl;
      exit( 0 );
      }
   }
